unit LDN168u1;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs, StdCtrls,Clipbrd;

const
   skVer='1 Sept 20';//LDN168 started 27 Aug 20
      //(Note of interest only to author- Was for Books Inventory,
      //   in particular "where is it?" tracking.)

   //This code presented in http://sheepdogguides.com/lut/lt4Nf-Barcode.htm

   skLaTxtChkCodeDataCaption='Data in barcode   /   checksum character. '+
         '(Above this line)';

   (*Built mostly from information in...
      https://en.wikipedia.org/wiki/Code_128#Check_digit_calculation

   UTF-8 character encoding is used thoughout. (Lazarus "speaks" UTF-8
      without any special measures.)
   A great deal of the coding uses, as is so widley the case, ASCII.
   However, a second encoding is importnat here, too. On the Wikipedia
   page, there is a big table showing how values convert between the
   two. There, the "second encoding" is called the "value" for the
   different characters. In most cases, the conversion can be handled
   with a mathematical rule. A few special cases exist. (This is also
   where UTF-8 is essential.)
   *)

   //Constant used here, derived from name used in Code128/
   //   UTF-8 for character in Latin-1 character set/
   //  (In comment) "Value" (And sometimes further notes).
   //The use of the characters for values 95-102 (incl) depends on their
   //   context except when they are being used as the Checksum character.
   ksVal95='Ã';//95;
   ksVal96='Ä';//96;
   ksVal97='Å';//97;
   ksVal98='Æ';//98;
   ksVal99='Ç';//99;
   //Look closely at the diacriticals... the codes ARE different!
   ksVal100='È';//100;
   ksVal101='É';//101;
   ksVal102='Ê';//102;
   ksStartCodeA='Ë';//103. This, if Ë, hardcodes the prgm to create CodeB barcodes
   ksStartCodeB='Ì';//104. This, if Ì, hardcodes the prgm to create CodeB barcodes
   ksStartCodeC='Í';//105. This, if Í, hardcodes the prgm to create CodeB barcodes
   ksStopCode='Î';// - Stop pattern of 4 bars, three spaces
   ksOutSideCode128Rogue='Ï';//This is the equivalent of what WOULD BE
      //the character for CodeB value 207- IF CODE128 //HAD// VALUE 207.
      //It doesn't. this is provided as a rogue value, for flagging attempts
      //to put use a value that Code128 does not provide for.

   //As it stands, this program "doesn't to" CodeA or CodeC encoding, but
   //the relevant constants are supplied for completeness

   (*Regarding "Code 128"/ "Code B":

      "Code 128" is an agreed method of putting data into a barcode.
         Everything here is "Code 128" happy.

      Code 128 has, "in it", three major variations... "Code"s A, B,
         and C. (And a few other clever things CAN be involved, too!)

      Everything here is assuming that you are "just" using Code B.
        (Some individual barcodes have a mix of, say, Code A and B!)

   ------------
   The character set is UTF-8. If an import of a CSV textfile to
      LibreOffice Calc is part of what you are doing, be sure to
      tell the import wizard to use UTF-8.

   Textpad.com offer an excellent Windows text editor, on very
      terms... full trial, not time limited, available. It
      seems to work in UTF-8, or an equivalent, "out of the box".

   Here are the ALL the characters allowed in the data portion of
      a barcode- when you are using Code128, sub-code B throughout
      ... which is, as I said, the only thing this code attempts.

   (space) ! " # $ % & ' ( ) * + , - . /
   0 1 ... 7 8 9 : ; < = > ? @ A-Z [ \ ] ^ _ ` a-z { | |

   (Certain international characters, e.g. "£" can NOT go in the
      data portion of the barcode.)

   The set just listed in full is merely all the ASCII characters with
   decimal values >31 and < 127. (Exclusive. Not ">=" or "<=")

   -------------
   Test data...
   The following is the complete string of characters to be
   rendered in a "Code 128" font, in order to present PJJ123C
   in a barcode...

     ÌPJJ123CWÎ

   Ì: Start CodeB
   PJJ123C: Data encoded inside the barcode
   W: Checkum
   Î:Stop Code

   ---------------------
   You will need a barcode scanner!

   But you CAN use a cheap USB webcam as your "barcode scanner".
   You can even use it to scan images of barcodes on
   your computer screen!

   There are MANY "all singing, all dancing, run Amazon or
   similar" "barcode programs" out there. If you just want
   something SIMPLE that WILL read a code 128 barcode, with
   a webcam, try...

   Freeware for Windows: Try either...

   1) Katanshi bar code reader.  It seemed the simplest.

   Now... Firefox objected when I tried to visit...
      http:\\katanshi.com
   ... but I was able to download it (and the download passed an eSet malware
      scan) from...
      https://download.cnet.com/Katanshi-Barcode-Reader/3001-2067_4-75221041.html

   --- alternatively....

   2) the scanner-by-webcam software from...

   https://bytescout.com/products/developer/barcodereadersdk/
                                       bytescoutbarcodereadersdk.html

   (You have to dig a little to find the freeware one, but I found this polished,
   quick, reliable. But remember to click the "Start" button in the "live view"
   window (!) Oh. And for this and the first one, do tell it you'll be looking
   at Code 128 barcodes.)
   *)


type

  { TLDN168f1 }

  TLDN168f1 = class(TForm)
    buDoIt: TButton;
    buTstCalcValue: TButton;
    buQuit: TButton;
    buReqs: TButton;
    eDataToGoToBC: TEdit;
    laClickMe: TLabel;
    laTxtFullContents: TLabel;
    laTxtLinkToTut: TLabel;
      //"buDoItClick" was made (via Object Inspector) the
      //handler for eDataToGoToBC's OnEditingDone event.
    laSampleTxt2: TLabel;
    laSampleTxt1: TLabel;
    laTxtChkCodeData: TLabel;
    laTxtDataToGoIn: TLabel;
    laTxtPrgmSrc: TLabel;
    laFullBCContents: TLabel;
    laBarcode: TLabel;
    laErrorMsgs: TLabel;
    laBCData: TLabel;
    laDebugRoutinesOutput: TLabel;
    laChkChar: TLabel;
    laTxtPrgmSrc1: TLabel;
    procedure buDoItClick(Sender: TObject);
    procedure buReqsClick(Sender: TObject);
    procedure buTstCalcValueClick(Sender: TObject);
    procedure buQuitClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure laTxtLinkToTutClick(Sender: TObject);

  private
    bTmp,bValReturned:byte;
    boTmp:boolean;
    sTmp:string;
    chTmp:char;
    function bGenChecksumChar(sBCData:string):byte;
    function bCharValOrErrCode(sToL:string):byte;
    function sCharFrmVal(bToConvert:byte):string;
    function boValidUSKeyboardCharacter(chCheckThis:char):boolean;
    function sSeeThatOnlyCodeableCharactersAskedFor(sToCode:string):string;
    procedure DoIt;

  public

  end;

var
  LDN168f1: TLDN168f1;

implementation

{$R *.lfm}

{ TLDN168f1 }

procedure TLDN168f1.FormCreate(Sender: TObject);
begin
  LDN168f1.caption:='LDN168-Barcode generator. Vers:'+skVer;
  application.title:='LDN168';
  top:=100;
  Left:=140;

  //These "could" be done in design... but that makes working on
  //  the form difficult.
  laTxtChkCodeData.caption:=skLaTxtChkCodeDataCaption;
  laChkChar.caption:='';
  laBCData.caption:='';
  laErrorMsgs.caption:='';
  laTxtFullContents.caption:='';
  buDoIt.caption:='Convert to Code 128 (B) barcode';

  //"Show" the following, if you want to use them for debugging...
      buTstCalcValue.hide;
      laDebugRoutinesOutput.hide;//Displays messages from buTstCalcValueClick

  DoIt;//initialize display, with barcode, etc, for whatever has been
       //  loaded into eDataToGoToBC.text

end;//FormCreate

procedure TLDN168f1.laTxtLinkToTutClick(Sender: TObject);
begin
 Clipboard.AsText:=laTxtLinkToTut.caption;
 showmessage('URL copied to clipboard');
end;

///////NEEDS EDIT...
procedure TLDN168f1.DoIt;
//The "top" procedure of the heart of this program.
var sTo,sErr,sBCData,sChkChar:string;
begin
   sErr:='0';//"Code" for "no errors seen yet.
   //Value in sErr only of interest down to...
   //"if sErr<>'0' then begin"
   //(It may change again before we get there, though.)

  sBCData:=eDataToGoToBC.text;
  //sBCData:='PJJ123C';//De-rem, if you wish, for debugging

  //---
  //What the next line does could be done more "efficiently"
  //elsewhere, but not as transparently. The time wasted will
  //not be perceptible by the human using the program.
  //Doing it here also helps keep the flowchart simple.
  sErr:=sSeeThatOnlyCodeableCharactersAskedFor(sBCData);

  if sErr='0' then bValReturned:=bGenChecksumChar(sBCData);
  //showmessage(inttostr(bValReturned));

  if bValReturned>199 then sErr:='Error arose trying to convert '+
       'character to value';

  if sErr<>'0' then begin
      //There was a problem in generating barcode. Details
      // of the most recent problem will be in sErr.
      laErrorMsgs.caption:=sErr;
      laFullBCContents.caption:='';
      laBarcode.caption:='';
      laBCData.caption:='';
      laChkChar.caption:='';
      laTxtChkCodeData.caption:='';
      end//no ; here
    else begin//... Proceed with barcode generation. No error yet.
      //This "begin" is the start of large block (20829a). It ends
      //where you see "**** End of large "ELSE" block (20829a)"

  sChkChar:=sCharFrmVal(bValReturned);
  laTxtChkCodeData.caption:=skLaTxtChkCodeDataCaption;

  //Assemble the finished barcode from the now-ready "bits"...
  sTo:=ksStartCodeB+sBCData+sChkChar+ksStopCode;

  //Put barcode and useful bits on screen...
    laTxtFullContents.caption:='Below:  ALL of the characters in the barcode:';
    laFullBCContents.caption:=sTo; //plaintext
    laBarcode.caption:=sTo; //as barcode... requires code128 font on system
    laBCData.caption:=sBCData;  //to go with barcode on label
    laChkChar.caption:=sChkChar;
    //Continually RE-set this, as it will be erased if there's a problem
    //  during a prior convert attempt.
    laTxtChkCodeData.caption:=skLaTxtChkCodeDataCaption;
    laErrorMsgs.caption:='';
  end;//**** End of large "ELSE" block (20829a)
end;//DoIt

function TLDN168f1.bGenChecksumChar(sBCData:string):byte;
(* This routine calculates, by the Code 128 rules... see the
Wikipeida article cited at the top of this... a checksum for
the start character and all of the data portion of the
barcode being generated.

  Test data: Pass this PJJ123C, and it should return ord('W')
    as long as you have left the rest of the program
    calculating strings for Code 128, scheme "B" barcodes.
*)
var iLoopCount,iCS:integer;
begin
  //sErrL:='Error codes- to be implemented';
  //sToL:=sToL;
  iCS:=104;//Why? Because 104 is the "value" of the "Start Code B",
     //character which is at the start of the barcode, and is to
     //be included in determining the checksum.
     //If you are trying to compare what you see on the Wikipedia
     //page, the "PJJ123C" example, with this, REMEMEMBER
     //AND ALLOW FOR THE FACT THAT WP IS ENCODING the data
     //into a Code *A* barcode. The only difference,
     //though, is that we use a different "start" code,
     //"value", 104 instead of 103, which, because of where
     //it comes in the string, means that our final value
     //for the checksum is higher by one than the value given
     //in the example. I.e. we get 55 for PJJ123C (translates
     //to "W", whereas the example gets 54, "V".

  for iLoopCount:=1 to length(sBCData) do begin
    //showmessage(sBCData[iLoopCount]);//De-rem if you are debugging
    //showmessage(inttostr(bCharValOrErrCode(sBCData[iLoopCount])));
    iCS:=iCS+(bCharValOrErrCode(sBCData[iLoopCount])*iLoopCount);
    //I have not tested for the return of a rogue value at this point
    //  because sufficient... famous last words... tests were done
    //  before we reached this point.
  end;//for...

  //showmessage('iCS before mod 103: '+inttostr(iCS));
  //showmessage('iCS after mod 103: '+inttostr(iCS mod 103));

  result:=iCS mod 103;//Why 103? 'Cause that's what Code128-B calls for.
end;//bGenChecksumChar

function TLDN168f1.bCharValOrErrCode(sToL:string):byte;
(*Calculates the "Code 128 / Code B" "value" of a character. ("Value" as
   used by standard Code 128 checksum calculations.)

   Returns a rogue value (anything >199) if an error encountered.

The following documentation is far longer than it needs to be of you only
used the "Code B" scheme, which is a standard part of Code 128. I leave
it as it is because if you want to expand this code to cover the "A"
and "C" schemes provided for within Code 128, you may find it useful.

There's nothing from here down to the return to code that you particulary
need to worry about unless you want to understand the details of the Code
128 system.

----
In code schemes B and A (but not C)..
Char    [space] !"#$%&'()*+,-./ 012...89:;... @ABC...Z
Value*     0 ...             15 16...                58
ASCII    32,33,34,35...      47 48... and they both just keep rising, to 90 for "Z".
      * "Value" is that as reckoned by "Code B" designers

After ...XYZ come the characters: "[\]^_"
                      values: 59-63
                      ASCII: 91-95

THEN Code A starts to be different from
  Code B (Code C has been different long since).

But the good news is that the simple "ASCII-32" continues valid.

64 (value) goes with ASCII 96 (the "backwards apostrophe") 96,

And then we are away again...

Char     a     b    c    d...     z   {   |   }   ~
Value*   65   66   67   68       90  91  92  93  94
ASCII    97   98   99  100      122 123 124 125 126
      * "Value" is that as reckoned by "Code B" designers

VALUES 95-106 are assigned to esoteric things I doubt someone
*getting started* with turning data into barcodes will find
necessary.

Values above 106 were not assigned, when Code 128 was being
designed... and so the bigger program this is part of treats
the return of such values as rogue value messengers saying
"something went wrong inside bCharValOrErrCode".
*)

var bValueL, bAsciiEquiv:byte;
begin
  //Reminders: ord('A')=65 and ord('a')=97
  bValueL:=200;//Default to "general" error
  bAsciiEquiv:=ord(sToL[1]);//Gives you first char of string
  if ((bAsciiEquiv>31) and
      (bAsciiEquiv<127)) then begin
           bValueL:=bAsciiEquiv-32
           end// no ; here
        else begin
           bValueL:=201;//Error code: character outside of
           //range allowed for Scheme B in Code 128.
           end;//else
  result:=bValueL;
end;//...bCharValOrErrCode


function TLDN168f1.sCharFrmVal(bToConvert:byte):string;
(*The heart of this is so simple because it is ONLY suitable
  for generating the right "Code 128" characters for scheme B barcodes.

  But the coding so far should allow extension for schemes A and C.

  See documentation in bCharValOrErrCode for details.

  The formula derives from the following case...
     55 should yield "W".
     The ASCII for W is 87.

  This doesn't have error trapping because it is intended only
     for use in a program where a bad Value will not be passed to it.

  When you see the results of some of the ksVal constants, be sure
     to look closely at the diacriticals... the characters ARE different!

  Beware: Our computers are "being helpful" again.

  In the Olden Days, it took three BYTES to display three characters.
  "ABC" arose from 65, 66 and 67 (decimal) being somewhere inside the
  computer. Broadly speaking, that is still true. // However, when we
  wanted "more", it came at a cost. We wanted, for instance to be able
  to display "©", the copyright //symbol//. To put that in the context
  I used before, A©C CAN be shown on our output devices... but it takes
  "UTF-8". And, in a system that "understands UTF-8... as most do today,
  "behind" A©C there are FOUR bytes stored somewhere: 65,194,169,67
 *)
begin

  case bToConvert of
     0..94:sTmp:=chr(bToConvert+32);//Deals with most cases
     95:sTmp:=ksVal95;
     96:sTmp:=ksVal96;
     97:sTmp:=ksVal97;
     98:sTmp:=ksVal98;
     99:sTmp:=ksVal99;
     100:sTmp:=ksVal100;
     101:sTmp:=ksVal101;
     102:sTmp:=ksVal102;
     //Only values 0-102 should be presented, because the
     //values needing "translating" should all have come
     //from x MOD 103.
  else //and yes, there SHOULD be a ; at end of previous line!
 //    sTmp:=ksOutSideCode128Rogue;//Ï -A code BEYOND those used by
 //       Code 128... a rogue value to say "Character sent to sCharFrmVal
 //       was not valid in Code 128"
  end;
  result:=sTmp;
end;//sCharFrmVal

function TLDN168f1.boValidUSKeyboardCharacter(chCheckThis:char):boolean;
//This is about catching, say, a "£" in the data to be put in the
//barcode. The £ is not provided for in Code 128-B. Happily, even though
//you can put a "£" in an edit box, within Lazarus, that will be seen
//as two bytes. ("ASCII" 194+163). The valid-for-Code 128-B characters
//have ASCII codes BETWEEN (not including) 31 and 127, making the test
//we need here easy to code.
begin
  boTmp:=false;
  bTmp:=ord(chCheckThis);
  if (bTmp>31)and(bTmp<127) then boTmp:=true;
  result:=boTmp;
end;//of boValidUSKeyboardCharacter

function TLDN168f1.sSeeThatOnlyCodeableCharactersAskedFor
                                          (sToCode:string):string;
//Returns '0', as string, if no problem found. Or a human-friendly
//   string, if an error is found.
var iLoopCount:integer;
begin
  sTmp:='0';//Hope for the best! But check...
    for iLoopCount:=1 to length(sToCode) do begin
      if (boValidUSKeyboardCharacter(sToCode[iLoopCount])=false) then begin
        sTmp:='Code 128 (scheme B) cannot put '+
               'one of those characters in a barcode';
        //Adding WHICH character was uncodeable was more trouble than
        //  I wanted to go to. The "uncodeable character" may well
        //  be represented by TWO (or more) bytes in the data stream,
        //  and at this stage in the process we are considering the
        //  stream one byte at a time.
      end;//if...
    end;//for...
  result:=sTmp;
end;

procedure TLDN168f1.buTstCalcValueClick(Sender: TObject);
//A routine for use during code development
begin
   bValReturned:=bCharValOrErrCode(eDataToGoToBC.text);
   if bValReturned<200 then begin
     laDebugRoutinesOutput.Caption:=inttostr(bValReturned);
     end//no ; here
    else begin
      laDebugRoutinesOutput.Caption:='Error seen during '+
           'bValFrmCharOr200plusForError';
    end;//else
end;//buTstCalcValueClick

(*Boring bits from here to end... *)
procedure TLDN168f1.buQuitClick(Sender: TObject);
begin
  application.terminate;
end;

procedure TLDN168f1.buDoItClick(Sender: TObject);
begin
  DoIt;
end;

procedure TLDN168f1.buReqsClick(Sender: TObject);
begin
  showmessage('To use this program, your computer will need'+chr(13)+
    'a font for the Code128 barcode characters. I got mine (free)'+chr(13)+
    'from https://www.wfonts.com');
end;//buReqsClick

end.

